/** Interface VRP simplifie pour ALTA. Initialisation par Madeleine.
 */

#include "VRP.h"



/* ** Structure des messages de VRP ALTA
 *
 * ------------------------------------------------------------------------
 * |      En-tte         |               Donnes                         |
 * | (garanti non-perdu)  | 'n_elm' lments de taille 'size_elm' chacun  |
 * |                      |      (avec tolrance de pertes)               |
 * ------------------------------------------------------------------------
 * ^                      ^                                               ^
 * 0                   offset                                     offset+n_elm*size_elm
 *
 */



/** Paramtres spcifiant la tolrance de perte (metteur).
 * S'appliquent uniquement  la partie "donnes" d'un message.
 */
struct vrp_alta_loss_parameters_s
{
  int max_loss_rate;    /**< taux global de pertes autorises (en %, de 0%  100%) */
  int max_cons_loss;    /**< plus grosse perte conscutive autorise */
  int min_refresh_rate; /**< nombre d'itrations conscutives autorises sans rafraichissement  */
};
typedef struct vrp_alta_loss_parameters_s*vrp_alta_loss_parameters_t;

static const struct vrp_alta_loss_parameters_s vrp_alta_reliable =
  {
    .max_loss_rate = 0,
    .max_cons_loss = -1,
    .min_refresh_rate = -1
  };

/** Feedback sur les pertes relles (ct rcepteur)
 */
struct vrp_alta_loss_feedback_s
{
  char*losses_array; /**< tableaux des pertes effectives. losses_array[i] == 1 <==> element #i perdu. */
  int size; /**< taille du tableau ; normalement, == n_elm */
};
typedef struct vrp_alta_loss_feedback_s*vrp_alta_loss_feedback_t;


/* ********************************************************* */
/* API publique */

/** Connexion des canaux VRP.  appeler *aprs* avoir initialis Madeleine et
 *  dmarr le thread de rception.
 */
void vrp_alta_connect(void);


/** Envoi d'un message, dispos selon la structure de message donne en figure.
 */
void vrp_alta_send(int dest_rank,
		   const void*data, int offset, int size_elm, int n_elm,
		   const struct vrp_alta_loss_parameters_s*loss_params);

/** Rception (bloquante) d'un message, selon la structure donne en figure.
 * 'data' est allou par cette fonction
 * 'losses' est rempli par cette fonction
 */
void vrp_alta_recv(void**data, int offset, int size_elm, int*n_elm,
		   vrp_alta_loss_feedback_t losses);

/** Nettoyage aprs une rception. Dsalloue le buffer de donnes et de
 * feedback de pertes proprement.  appeler un (et une seule) fois aprs
 * la rception de chaque message.
 */
static inline void vrp_alta_cleanup(void**_data, vrp_alta_loss_feedback_t losses)
{
  if(losses->losses_array)
    {
      TBX_FREE(losses->losses_array);
      losses->losses_array = NULL;
      losses->size = 0;
    }
  TBX_FREE(*_data);
  *_data = NULL;
}



/* ********************************************************* */
/* fonctions internes  Mad/VRP */

void vrp_alta_set_incoming(vrp_incoming_t incoming, int rank);
void vrp_alta_set_outgoing(vrp_outgoing_t outgoing, int rank);

/* ********************************************************* */


