
/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <stdlib.h>

#section types
typedef struct __marcel_attr_s marcel_attr_t;

#section macros
#define MARCEL_MAXNAMESIZE 16

#section structures

#if defined(LINUX_SYS) || defined(DARWIN_SYS)
#include <sched.h>
#else // AD: fake Linux-like structures on other systems
#include <sys/signal.h>
struct sched_param
{
  int __sched_priority;
};
#ifndef GNU_SYS
typedef sigset_t __sigset_t;
#endif
#endif

#include <sys/types.h> /* pour size_t */
#depend "marcel_sched_generic.h[]"

/* Attributes for threads.  */
struct __marcel_attr_s
{
  /* begin of pthread */
  int __detachstate;
  int __schedpolicy;
  struct sched_param __schedparam;
  int __inheritsched;
  int __scope;
  size_t __guardsize;
  int __stackaddr_set;
  void *__stackaddr;
  size_t __stacksize;
  /* end of pthread */

  /* marcel attributes */
  //unsigned stack_size;
  //char *stack_base;
  //int /*boolean*/ detached;
  unsigned user_space;
  /*boolean*/int immediate_activation;
  unsigned not_migratable;
  unsigned not_deviatable;
  int not_preemptible;
  //int sched_policy;
  /*boolean int rt_thread; On utilise la priorit maintenant */
  marcel_vpmask_t vpmask;
  int flags;
  char name[MARCEL_MAXNAMESIZE];
  marcel_sched_attr_t sched;
};


#section macros
// MARCEL_CREATE_JOINABLE
#depend "marcel_threads.h[types]"
// MARCEL_SCHED_OTHER
// MARCEL_SCHED_ATTR_INITIALIZER
#define MARCEL_SCHED_INTERNAL_INCLUDE
#depend "scheduler/marcel_sched.h[marcel_macros]"
// MARCEL_VPMASK_EMPTY
#depend "marcel_sched_generic.h[macros]"

#define MARCEL_ATTR_INITIALIZER { \
  .__detachstate= MARCEL_CREATE_JOINABLE, \
  .__schedpolicy= MARCEL_SCHED_OTHER, \
  .__schedparam= {0,}, \
  .__inheritsched= 0, \
  .__scope= 0, \
  .__guardsize= 0, \
  .__stackaddr_set= 0, \
  .__stackaddr= NULL, \
  .user_space= 0, \
  .immediate_activation= FALSE, \
  .not_migratable= 1, \
  .not_deviatable= 0, \
  .not_preemptible= 0, \
  .vpmask= MARCEL_VPMASK_EMPTY, \
  .flags= 0, \
  .name= "user_task", \
  .sched= MARCEL_SCHED_ATTR_INITIALIZER, \
}

/* detachstate */
/*#define MARCEL_CREATE_JOINABLE    FALSE
#define MARCEL_CREATE_DETACHED    TRUE
*/

/* realtime */
#define MARCEL_CLASS_REGULAR      FALSE
#define MARCEL_CLASS_REALTIME     TRUE

#section functions
#depend "marcel_utils.h[types]"

int marcel_attr_init(marcel_attr_t *attr) __THROW;
#define marcel_attr_destroy(attr_ptr)	0

int marcel_attr_setstacksize(marcel_attr_t *attr, size_t stack) __THROW;
int marcel_attr_getstacksize(__const marcel_attr_t *attr, size_t *stack) __THROW;

int marcel_attr_setstackaddr(marcel_attr_t *attr, void *addr) __THROW;
int marcel_attr_getstackaddr(__const marcel_attr_t *attr, void **addr) __THROW;

int marcel_attr_setdetachstate(marcel_attr_t *attr, boolean detached) __THROW;
int marcel_attr_getdetachstate(__const marcel_attr_t *attr, boolean *detached) __THROW;

int marcel_attr_setuserspace(marcel_attr_t *attr, unsigned space);
int marcel_attr_getuserspace(__const marcel_attr_t *attr, unsigned *space);

int marcel_attr_setactivation(marcel_attr_t *attr, boolean immediate);
int marcel_attr_getactivation(__const marcel_attr_t *attr, boolean *immediate);

int marcel_attr_setmigrationstate(marcel_attr_t *attr, boolean migratable);
int marcel_attr_getmigrationstate(__const marcel_attr_t *attr, boolean *migratable);

int marcel_attr_setdeviationstate(marcel_attr_t *attr, boolean deviatable);
int marcel_attr_getdeviationstate(__const marcel_attr_t *attr, boolean *deviatable);

int marcel_attr_setschedpolicy(marcel_attr_t *attr, int policy) __THROW;
int marcel_attr_getschedpolicy(__const marcel_attr_t *attr, int *policy) __THROW;

int marcel_attr_setrealtime(marcel_attr_t *attr, boolean realtime);
int marcel_attr_getrealtime(__const marcel_attr_t *attr, boolean *realtime);

int marcel_attr_setvpmask(marcel_attr_t *attr, marcel_vpmask_t mask);
int marcel_attr_getvpmask(__const marcel_attr_t *attr, marcel_vpmask_t *mask);

int marcel_attr_setname(marcel_attr_t *attr, const char *name);
int marcel_attr_getname(__const marcel_attr_t *attr, char *name, size_t n);

int marcel_attr_setpreemptible(marcel_attr_t *attr, int preemptible);
int marcel_attr_getpreemptible(__const marcel_attr_t *attr, int *preemptible);

// only for internal use
int marcel_attr_setflags(marcel_attr_t *attr, int flags);
int marcel_attr_getflags(__const marcel_attr_t *attr, int *flags);

#section marcel_variables
extern marcel_attr_t marcel_attr_default;

